#!/usr/bin/env python

__version__ = '1.0.9'
__author__ = 'vin8tech'

import os
import sys

# make indicators available as pandas extensions
# import kinetick.lib.indicators as indicators
from . import *

# check min, python version
from . import instrument
from .lib import indicators
from .lib.brokers import Webull
from .lib.brokers import Zerodha

if sys.version_info < (3, 4):
    raise SystemError("Kinetick requires Python version >= 3.4")

path = {
    "library": os.path.dirname(os.path.realpath(__file__)),
    "caller": os.path.dirname(os.path.realpath(sys.argv[0]))
}

__all__ = [
    'blotter',
    'Zerodha',
    'bot',
    'models',
    'tests',
    'risk_assessor',
    'enums',
    'lib',
    'instrument'
    'algo',
    'broker',
    'Webull',
    'utils',
    'indicators',
    'factory'
    'strategies'
    'path'
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# QTPyLib: Quantitative Trading Python Library
# https://github.com/ranaroussi/qtpylib
#
# Copyright 2016-2018 Ran Aroussi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

_BARS_COLSMAP = {
    'open': 'open',
    'high': 'high',
    'low': 'low',
    'close': 'close',
    'volume': 'volume',
    'opt_price': 'opt_price',
    'opt_underlying': 'opt_underlying',
    'opt_dividend': 'opt_dividend',
    'opt_volume': 'opt_volume',
    'opt_iv': 'opt_iv',
    'opt_oi': 'opt_oi',
    'opt_delta': 'opt_delta',
    'opt_gamma': 'opt_gamma',
    'opt_vega': 'opt_vega',
    'opt_theta': 'opt_theta'
}
_TICKS_COLSMAP = {
    'bid': 'bid',
    'bidsize': 'bidsize',
    'ask': 'ask',
    'asksize': 'asksize',
    'last': 'last',
    'lastsize': 'lastsize',
    'opt_price': 'opt_price',
    'opt_underlying': 'opt_underlying',
    'opt_dividend': 'opt_dividend',
    'opt_volume': 'opt_volume',
    'opt_iv': 'opt_iv',
    'opt_oi': 'opt_oi',
    'opt_delta': 'opt_delta',
    'opt_gamma': 'opt_gamma',
    'opt_vega': 'opt_vega',
    'opt_theta': 'opt_theta'
}

# ---------------------------------------------


def validate_columns(df, kind="BAR", raise_errors=True):
    global _TICKS_COLSMAP, _BARS_COLSMAP
    # validate columns
    if "asset_class" not in df.columns:
        if raise_errors:
            raise ValueError('Column asset_class not found')
        return False

    is_option = "OPT" in list(df['asset_class'].unique())

    colsmap = _TICKS_COLSMAP if kind == "TICK" else _BARS_COLSMAP

    for el in colsmap:
        col = colsmap[el]
        if col not in df.columns:
            if "opt_" in col and is_option:
                if raise_errors:
                    raise ValueError('Column %s not found' % el)
                return False
            elif "opt_" not in col and not is_option:
                if raise_errors:
                    raise ValueError('Column %s not found' % el)
                return False
    return True


# =============================================
# data analyze methods
# =============================================

def analyze_portfolio(file):
    """ analyze portfolio (TBD) """
    pass



import pandas as pd
import numpy as np


# Source for tech indicator : https://github.com/arkochhar/Technical-Indicators/blob/master/indicator/indicators.py
def EMA(df, base, target, period, alpha=False):
    """
    Function to compute Exponential Moving Average (EMA)

    Args :
        df : Pandas DataFrame which contains ['date', 'open', 'high', 'low', 'close', 'volume'] columns
        base : String indicating the column name from which the EMA needs to be computed from
        target : String indicates the column name to which the computed data needs to be stored
        period : Integer indicates the period of computation in terms of number of candles
        alpha : Boolean if True indicates to use the formula for computing EMA using alpha (default is False)

    Returns :
        df : Pandas DataFrame with new column added with name 'target'
    """

    con = pd.concat([df[:period][base].rolling(window=period).mean(), df[period:][base]])

    if (alpha == True):
        # (1 - alpha) * previous_val + alpha * current_val where alpha = 1 / period
        df[target] = con.ewm(alpha=1 / period, adjust=False).mean()
    else:
        # ((current_val - previous_val) * coeff) + previous_val where coeff = 2 / (period + 1)
        df[target] = con.ewm(span=period, adjust=False).mean()

    df[target].fillna(0, inplace=True)
    return df


def ATR(df, period, ohlc=['open', 'high', 'low', 'close']):
    """
    Function to compute Average True Range (ATR)

    Args :
        df : Pandas DataFrame which contains ['date', 'open', 'high', 'low', 'close', 'volume'] columns
        period : Integer indicates the period of computation in terms of number of candles
        ohlc: List defining OHLC Column names (default ['Open', 'High', 'Low', 'Close'])

    Returns :
        df : Pandas DataFrame with new columns added for
            True Range (TR)
            ATR (ATR_$period)
    """
    atr = 'ATR_' + str(period)

    # Compute true range only if it is not computed and stored earlier in the df
    if not 'TR' in df.columns:
        df['h-l'] = df[ohlc[1]] - df[ohlc[2]]
        df['h-yc'] = abs(df[ohlc[1]] - df[ohlc[3]].shift())
        df['l-yc'] = abs(df[ohlc[2]] - df[ohlc[3]].shift())

        df['TR'] = df[['h-l', 'h-yc', 'l-yc']].max(axis=1)

        df.drop(['h-l', 'h-yc', 'l-yc'], inplace=True, axis=1)

    # Compute EMA of true range using ATR formula after ignoring first row
    EMA(df, 'TR', atr, period, alpha=True)

    return df


def SuperTrend(df, period=7, multiplier=3, ohlc=['open', 'high', 'low', 'close']):
    """
    Function to compute SuperTrend

    Args :
        df : Pandas DataFrame which contains ['date', 'open', 'high', 'low', 'close', 'volume'] columns
        period : Integer indicates the period of computation in terms of number of candles
        multiplier : Integer indicates value to multiply the ATR
        ohlc: List defining OHLC Column names (default ['Open', 'High', 'Low', 'Close'])

    Returns :
        df : Pandas DataFrame with new columns added for
            True Range (TR), ATR (ATR_$period)
            SuperTrend (ST_$period_$multiplier)
            SuperTrend Direction (STX_$period_$multiplier)
    """

    ATR(df, period, ohlc=ohlc)
    atr = 'ATR_' + str(period)
    st = 'ST'  # + str(period) + '_' + str(multiplier)
    stx = 'STX'  # + str(period) + '_' + str(multiplier)

    """
    SuperTrend Algorithm :

        BASIC UPPERBAND = (HIGH + LOW) / 2 + Multiplier * ATR
        BASIC LOWERBAND = (HIGH + LOW) / 2 - Multiplier * ATR

        FINAL UPPERBAND = IF( (Current BASICUPPERBAND < Previous FINAL UPPERBAND) or (Previous Close > Previous FINAL UPPERBAND))
                            THEN (Current BASIC UPPERBAND) ELSE Previous FINALUPPERBAND)
        FINAL LOWERBAND = IF( (Current BASIC LOWERBAND > Previous FINAL LOWERBAND) or (Previous Close < Previous FINAL LOWERBAND)) 
                            THEN (Current BASIC LOWERBAND) ELSE Previous FINAL LOWERBAND)

        SUPERTREND = IF((Previous SUPERTREND = Previous FINAL UPPERBAND) and (Current Close <= Current FINAL UPPERBAND)) THEN
                        Current FINAL UPPERBAND
                    ELSE
                        IF((Previous SUPERTREND = Previous FINAL UPPERBAND) and (Current Close > Current FINAL UPPERBAND)) THEN
                            Current FINAL LOWERBAND
                        ELSE
                            IF((Previous SUPERTREND = Previous FINAL LOWERBAND) and (Current Close >= Current FINAL LOWERBAND)) THEN
                                Current FINAL LOWERBAND
                            ELSE
                                IF((Previous SUPERTREND = Previous FINAL LOWERBAND) and (Current Close < Current FINAL LOWERBAND)) THEN
                                    Current FINAL UPPERBAND
    """

    # Compute basic upper and lower bands
    df['basic_ub'] = (df[ohlc[1]] + df[ohlc[2]]) / 2 + multiplier * df[atr]
    df['basic_lb'] = (df[ohlc[1]] + df[ohlc[2]]) / 2 - multiplier * df[atr]

    # Compute final upper and lower bands
    df['final_ub'] = 0.00
    df['final_lb'] = 0.00
    for i in range(period, len(df)):
        df['final_ub'].iat[i] = df['basic_ub'].iat[i] if df['basic_ub'].iat[i] < df['final_ub'].iat[i - 1] or \
                                                         df[ohlc[3]].iat[i - 1] > df['final_ub'].iat[i - 1] else \
            df['final_ub'].iat[i - 1]
        df['final_lb'].iat[i] = df['basic_lb'].iat[i] if df['basic_lb'].iat[i] > df['final_lb'].iat[i - 1] or \
                                                         df[ohlc[3]].iat[i - 1] < df['final_lb'].iat[i - 1] else \
            df['final_lb'].iat[i - 1]

    # Set the Supertrend value
    df[st] = 0.00
    for i in range(period, len(df)):
        df[st].iat[i] = df['final_ub'].iat[i] if df[st].iat[i - 1] == df['final_ub'].iat[i - 1] and df[ohlc[3]].iat[
            i] <= df['final_ub'].iat[i] else \
            df['final_lb'].iat[i] if df[st].iat[i - 1] == df['final_ub'].iat[i - 1] and df[ohlc[3]].iat[i] > \
                                     df['final_ub'].iat[i] else \
                df['final_lb'].iat[i] if df[st].iat[i - 1] == df['final_lb'].iat[i - 1] and df[ohlc[3]].iat[i] >= \
                                         df['final_lb'].iat[i] else \
                    df['final_ub'].iat[i] if df[st].iat[i - 1] == df['final_lb'].iat[i - 1] and df[ohlc[3]].iat[i] < \
                                             df['final_lb'].iat[i] else 0.00

        # Mark the trend direction up/down
    df[stx] = np.where((df[st] > 0.00), np.where((df[ohlc[3]] < df[st]), 'down', 'up'), np.NaN)

    # Remove basic and final bands from the columns
    df.drop(['basic_ub', 'basic_lb', 'final_ub', 'final_lb'], inplace=True, axis=1)

    # df.fillna(0, inplace=True)
    return df



from .buy_low_sell_high import BuyLowSellHigh


def strategy():
    algo = BuyLowSellHigh(
        instruments=['SBIN', 'HDFC'],
        tick_window=100,
        bar_window=800,
        preload="1D",
        output="./orders.csv",
        timezone="Asia/Calcutta",
        # backtest=True,
        # start="2020-07-01 00:15:00",
        # backfill=True,
    )

    return algo



# Copyright 2021 vin8tech, vinay patil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from math import floor

from kinetick.bot import Bot
from kinetick.models import Position


class Borg:
    _shared_state = {}

    def __init__(self):
        self.__dict__ = self._shared_state


class RiskAssessor(Borg):
    def __init__(self, initial_capital=1000, max_trades=1, initial_margin=100,
                 risk2reward=1.0, risk_per_trade=100, **kwargs):
        Borg.__init__(self)

        self.capital = initial_capital
        self.initial_capital = initial_capital
        self.max_trades = max_trades
        self.available_margin = initial_margin
        self.initial_margin = initial_margin
        self.risk2reward = risk2reward
        self.risk_per_trade = risk_per_trade
        self.pnl = 0
        self.active_positions = []
        self.win_trades = 0
        self.loss_trades = 0

        if initial_capital < initial_margin:
            raise Exception("Capital is lower than available_margin")

        Bot().add_command_handler("report", self.availableMarginHandler, "Get report")
        Bot().add_command_handler("resetrms", self.reset, "Reset RMS values")

    def availableMarginHandler(self, update, context):
        msg = "```\n" \
              "Available margin: {:.2f} \n" \
              "Active Trades: {} \n" \
              "Available capital: {:.2f} \n" \
              "PNL: {:.2f} \n" \
              "Win Trades: {} \n" \
              "Loss Trades: {} \n" \
              "```".format(self.available_margin, len(self.active_positions),
                           self.capital, self.pnl, self.win_trades, self.loss_trades)
        update.message.reply_text(msg, parse_mode="MarkdownV2")

    @staticmethod
    def get_default_instance():
        return RiskAssessor(initial_capital=6000, max_trades=2, initial_margin=6000, risk2reward=1.3)

    def reset(self, update=None, ctx=None, **kwargs):
        self.capital = self.initial_capital
        self.available_margin = self.initial_margin
        self.active_positions.clear()
        self.pnl = 0
        self.win_trades = 0
        self.loss_trades = 0
        if update is not None:
            update.message.reply_text("RMS has been reset.")

    def _should_trade(self, entry_price, stop_loss, quantity=None):
        spread = abs(entry_price - stop_loss)
        spread = 5 * round(spread / 5, 2)
        maxsize = floor(max(1, int(self.risk_per_trade / spread)))
        maxsize = min(maxsize, int(self.capital / entry_price))
        _quantity = quantity or maxsize
        margin = _quantity * spread
        should_trade = True
        reason = None

        if self.available_margin <= 0:
            should_trade = False
            reason = "Insufficient margin"
        elif margin >= self.available_margin:
            should_trade = False
            reason = "Margin exceeds"
        elif len(self.active_positions) >= self.max_trades:
            should_trade = False
            reason = f'Exceeds max trade limit {self.max_trades}'
        elif (entry_price * _quantity) >= self.capital:
            should_trade = False
            reason = f'Insufficient Capital. required: {entry_price * _quantity} available: {self.capital}'
        elif quantity and int(quantity) > maxsize:
            should_trade = False
            reason = f'given quantity: {quantity} exceeds risk per trade: {self.risk_per_trade}, ' \
                     f'max allowed quantity {maxsize}'

        quantity = _quantity
        return should_trade, spread, quantity, reason

    def create_position(self, entry_price, stop_loss, quantity=None):
        """
        return trade if all the conditions are met
        :param quantity: int
        :param entry_price:
        :param stop_loss:
        :return: should_trade: Bool, position: Position
        """
        should_trade, spread, quantity, reason = self._should_trade(entry_price, stop_loss, quantity=quantity)
        if not should_trade:
            raise Exception(f'Trade can not be made voids risk parameters. {reason}')

        direction = "LONG" if stop_loss < entry_price else "SHORT"
        target = 5 * round((spread * self.risk2reward) / 5, 2)
        target = entry_price + target if direction == "LONG" else entry_price - target

        position = Position(_quantity=quantity, entry_price=entry_price, target=target,
                            stop=stop_loss, _direction=direction)

        return position

    # TODO make thread safe
    def enter_position(self, position):
        if position.entry_time is None:
            position.entry_time = datetime.now()
        should_trade, spread, quantity, reason = self._should_trade(position.entry_price, position.stop,
                                                                    quantity=position.quantity)
        if not should_trade:
            raise Exception(f'Trade can not be made voids rms parameters. {reason}')
        self.active_positions.append(position)
        self.available_margin = self.available_margin - spread * quantity
        self.capital = self.capital - (position.entry_price * quantity)
        return position

    # TODO make thread safe
    def exit_position(self, position):
        if position.exit_time is None:
            position.exit_time = datetime.now()

        pnl = position.pnl()

        # reclaim margin
        spread = abs(position.entry_price - position.stop)
        spread = 5 * round(spread / 5, 2)
        self.available_margin += spread * position.quantity + pnl
        self.available_margin = self.available_margin if self.available_margin < self.initial_margin \
            else self.initial_margin
        # reclaim the capital deployed in trade
        self.capital += (position.entry_price * position.quantity) + pnl
        self.pnl += pnl
        if pnl > 0:
            self.win_trades += 1
        else:
            self.loss_trades += 1

        index = self.active_positions.index(position)
        del self.active_positions[index]

        return position



from mongoengine import Document, StringField

CURRENCY = ('INR', 'USD')


class Symbol(Document):
    tickerId = StringField(max_length=50, required=True, primary_key=True)
    symbol = StringField(max_length=50, required=True, unique=True)
    sec_type = StringField(max_length=50, required=False)
    exchange = StringField(max_length=50, required=False)
    currency = StringField(max_length=3, choices=CURRENCY)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# QTPyLib: Quantitative Trading Python Library
# https://github.com/ranaroussi/qtpylib
#
# Copyright 2016-2018 Ran Aroussi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
from pandas import Series, DataFrame

from .supertrend import SuperTrend as supertrend

TALIB_MISSING = False

try:
    import talib
except ImportError:
    TALIB_MISSING = True
    raise ImportError("TA-Lib is not installed on this system!")

# =============================================
# check min, python version
if sys.version_info < (3, 4):
    raise SystemError("QTPyLib requires Python version >= 3.4")


# =============================================


# ---------------------------------------------
def _check_talib_presence():
    if TALIB_MISSING:
        raise ImportError("TA-Lib is not installed on this system!")


# ---------------------------------------------


def _extract_series(data):
    values = None

    if isinstance(data, Series):
        values = data.values
    else:
        if "last" in data.columns:
            values = data['last'].values
        elif "close" in data.columns:
            values = data['close'].values

    if values is None:
        raise ValueError(
            "data must be Pandas Series or DataFrame with a 'last' or 'close' column")

    return values


# ---------------------------------------------


def _extract_ohlc(data):
    if isinstance(data, DataFrame):
        if "open" in data.columns and "high" in data.columns \
                and "low" in data.columns and "close" in data.columns \
                and "volume" in data.columns:
            return data[['open', 'high', 'low', 'close', 'volume']].T.values

    raise ValueError("data must be Pandas with OLHC columns")


# ---------------------------------------------
# Overlap Studies
# ---------------------------------------------
def BBANDS(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.BBANDS(prices, **kwargs)


def DEMA(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.DEMA(prices, **kwargs)


def EMA(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.EMA(prices, **kwargs)


def HT_TRENDLINE(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.HT_TRENDLINE(prices, **kwargs)


def KAMA(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.KAMA(prices, **kwargs)


def MA(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MA(prices, **kwargs)


def MAMA(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MAMA(prices, **kwargs)


def MAVP(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MAVP(prices, **kwargs)


def MIDPOINT(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MIDPOINT(prices, **kwargs)


def MIDPRICE(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.MIDPRICE(phigh, plow, **kwargs)


def SAR(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.SAR(phigh, plow, **kwargs)


def SAREXT(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.SAREXT(prices, **kwargs)


def SMA(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.SMA(phigh, plow, **kwargs)


def T3(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.T3(prices, **kwargs)


def TEMA(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.TEMA(prices, **kwargs)


def TRIMA(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.TRIMA(prices, **kwargs)


def WMA(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.WMA(prices, **kwargs)


# ---------------------------------------------
# Momentum Indicators
# ---------------------------------------------
def ADX(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.ADX(phigh, plow, pclose, **kwargs)


def ADXR(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.ADXR(phigh, plow, pclose, **kwargs)


def APO(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.APO(prices, **kwargs)


def AROON(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.AROON(phigh, plow, **kwargs)


def AROONOSC(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.AROONOSC(phigh, plow, **kwargs)


def BOP(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.BOP(popen, phigh, plow, pclose, **kwargs)


def CCI(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CCI(phigh, plow, pclose, **kwargs)


def CMO(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.CMO(prices, **kwargs)


def DX(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.DX(phigh, plow, pclose, **kwargs)


def MACD(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MACD(prices, **kwargs)


def MACDEXT(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MACDEXT(prices, **kwargs)


def MACDFIX(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MACDFIX(prices, **kwargs)


def MFI(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, pvolume = _extract_ohlc(data)
    return talib.MFI(popen, phigh, plow, pclose, pvolume, **kwargs)


def MINUS_DI(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.MINUS_DI(phigh, plow, pclose, **kwargs)


def MINUS_DM(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.MINUS_DM(phigh, plow, **kwargs)


def MOM(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MOM(prices, **kwargs)


def PLUS_DI(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.PLUS_DI(phigh, plow, pclose, **kwargs)


def PLUS_DM(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.PLUS_DM(phigh, plow, **kwargs)


def PPO(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.PPO(prices, **kwargs)


def ROC(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.ROC(prices, **kwargs)


def ROCP(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.ROCP(prices, **kwargs)


def ROCR(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.ROCR(prices, **kwargs)


def ROCR100(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.ROCR100(prices, **kwargs)


def RSI(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.RSI(prices, **kwargs)


def STOCH(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.STOCH(phigh, plow, pclose, **kwargs)


def STOCHF(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.STOCHF(phigh, plow, pclose, **kwargs)


def STOCHRSI(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.STOCHRSI(prices, **kwargs)


def STOCHRSIS(data, **kwargs):
    """
    # this is the library function
    # >>> k, d = talib.STOCHRSI(c)

    # this produces the same result, calling STOCHF
    # >>> rsi = talib.RSI(c)
    # >>> k, d = talib.STOCHF(rsi, rsi, rsi)

    # you might want this instead, calling STOCH
    # >>> rsi = talib.RSI(c)
    # >>> k, d = talib.STOCH(rsi, rsi, rsi)
    Parameters:
        fastk_period: 5
        slowk_period: 3
        slowk_matype: 0
        slowd_period: 3
        slowd_matype: 0
    Outputs:
        slowk
        slowd
        :param data: pandas.df
    """
    _check_talib_presence()
    prices = RSI(data)
    return talib.STOCH(prices, prices, prices, **kwargs)


def TRIX(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.TRIX(prices, **kwargs)


def ULTOSC(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.ULTOSC(phigh, plow, pclose, **kwargs)


def WILLR(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.WILLR(phigh, plow, pclose, **kwargs)


# ---------------------------------------------
# Volume Indicators
# ---------------------------------------------
def AD(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, pvolume = _extract_ohlc(data)
    return talib.AD(popen, phigh, plow, pclose, pvolume, **kwargs)


def ADOSC(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, pvolume = _extract_ohlc(data)
    return talib.ADOSC(popen, phigh, plow, pclose, pvolume, **kwargs)


def OBV(data, **kwargs):
    _check_talib_presence()
    _, _, _, _, pvolume = _extract_ohlc(data)
    return talib.OBV(pvolume, **kwargs)


# ---------------------------------------------
# Cycle Indicators
# ---------------------------------------------
def HT_DCPERIOD(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.HT_DCPERIOD(prices, **kwargs)


def HT_DCPHASE(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.HT_DCPHASE(prices, **kwargs)


def HT_PHASOR(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.HT_PHASOR(prices, **kwargs)


def HT_SINE(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.HT_SINE(prices, **kwargs)


def HT_TRENDMODE(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.HT_TRENDMODE(prices, **kwargs)


# ---------------------------------------------
# Price Transform
# ---------------------------------------------
def AVGPRICE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.AVGPRICE(popen, phigh, plow, pclose, **kwargs)


def MEDPRICE(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.MEDPRICE(phigh, plow, **kwargs)


def TYPPRICE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.TYPPRICE(popen, phigh, plow, pclose, **kwargs)


def WCLPRICE(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.WCLPRICE(phigh, plow, pclose, **kwargs)


# ---------------------------------------------
# Volatility Indicators
# ---------------------------------------------
def ATR(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.ATR(phigh, plow, pclose, **kwargs)


def NATR(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.NATR(phigh, plow, pclose, **kwargs)


def TRANGE(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.TRANGE(phigh, plow, pclose, **kwargs)


# ---------------------------------------------
# Parrern Recognition
# ---------------------------------------------
def CDL2CROWS(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDL2CROWS(popen, phigh, plow, pclose, **kwargs)


def CDL3BLACKCROWS(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDL3BLACKCROWS(popen, phigh, plow, pclose, **kwargs)


def CDL3INSIDE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDL3INSIDE(popen, phigh, plow, pclose, **kwargs)


def CDL3LINESTRIKE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDL3LINESTRIKE(popen, phigh, plow, pclose, **kwargs)


def CDL3OUTSIDE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDL3OUTSIDE(popen, phigh, plow, pclose, **kwargs)


def CDL3STARSINSOUTH(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDL3STARSINSOUTH(popen, phigh, plow, pclose, **kwargs)


def CDL3WHITESOLDIERS(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDL3WHITESOLDIERS(popen, phigh, plow, pclose, **kwargs)


def CDLABANDONEDBABY(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLABANDONEDBABY(popen, phigh, plow, pclose, **kwargs)


def CDLADVANCEBLOCK(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLADVANCEBLOCK(popen, phigh, plow, pclose, **kwargs)


def CDLBELTHOLD(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLBELTHOLD(popen, phigh, plow, pclose, **kwargs)


def CDLBREAKAWAY(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLBREAKAWAY(popen, phigh, plow, pclose, **kwargs)


def CDLCLOSINGMARUBOZU(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLCLOSINGMARUBOZU(popen, phigh, plow, pclose, **kwargs)


def CDLCONCEALBABYSWALL(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLCONCEALBABYSWALL(popen, phigh, plow, pclose, **kwargs)


def CDLCOUNTERATTACK(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLCOUNTERATTACK(popen, phigh, plow, pclose, **kwargs)


def CDLDARKCLOUDCOVER(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLDARKCLOUDCOVER(popen, phigh, plow, pclose, **kwargs)


def CDLDOJI(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLDOJI(popen, phigh, plow, pclose, **kwargs)


def CDLDOJISTAR(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLDOJISTAR(popen, phigh, plow, pclose, **kwargs)


def CDLDRAGONFLYDOJI(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLDRAGONFLYDOJI(popen, phigh, plow, pclose, **kwargs)


def CDLENGULFING(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLENGULFING(popen, phigh, plow, pclose, **kwargs)


def CDLEVENINGDOJISTAR(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLEVENINGDOJISTRAR(popen, phigh, plow, pclose, **kwargs)


def CDLEVENINGSTAR(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLEVENINGSTAR(popen, phigh, plow, pclose, **kwargs)


def CDLGAPSIDESIDEWHITE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLGAPSIDESIDEWHITEITE(popen, phigh, plow, pclose, **kwargs)


def CDLGRAVESTONEDOJI(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLGRAVESTONEDOJII(popen, phigh, plow, pclose, **kwargs)


def CDLHAMMER(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLHAMMER(popen, phigh, plow, pclose, **kwargs)


def CDLHANGINGMAN(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLHANGINGMAN(popen, phigh, plow, pclose, **kwargs)


def CDLHARAMI(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLHARAMI(popen, phigh, plow, pclose, **kwargs)


def CDLHARAMICROSS(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLHARAMICROSS(popen, phigh, plow, pclose, **kwargs)


def CDLHIGHWAVE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLHIGHWAVE(popen, phigh, plow, pclose, **kwargs)


def CDLHIKKAKE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLHIKKAKE(popen, phigh, plow, pclose, **kwargs)


def CDLHIKKAKEMOD(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLHIKKAKEMOD(popen, phigh, plow, pclose, **kwargs)


def CDLHOMINGPIGEON(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLHOMINGPIGEON(popen, phigh, plow, pclose, **kwargs)


def CDLIDENTICAL3CROWS(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLIDENTICAL3CROWS(popen, phigh, plow, pclose, **kwargs)


def CDLINNECK(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLINNECK(popen, phigh, plow, pclose, **kwargs)


def CDLINVERTEDHAMMER(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLINVERTEDHAMMER(popen, phigh, plow, pclose, **kwargs)


def CDLKICKING(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLKICKING(popen, phigh, plow, pclose, **kwargs)


def CDLKICKINGBYLENGTH(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLKICKINGBYLENGTH(popen, phigh, plow, pclose, **kwargs)


def CDLLADDERBOTTOM(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLLADDERBOTTOM(popen, phigh, plow, pclose, **kwargs)


def CDLLONGLEGGEDDOJI(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLLONGLEGGEDDOJI(popen, phigh, plow, pclose, **kwargs)


def CDLLONGLINE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLLONGLINE(popen, phigh, plow, pclose, **kwargs)


def CDLMARUBOZU(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLMARUBOZU(popen, phigh, plow, pclose, **kwargs)


def CDLMATCHINGLOW(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLMATCHINGLOW(popen, phigh, plow, pclose, **kwargs)


def CDLMATHOLD(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLMATHOLD(popen, phigh, plow, pclose, **kwargs)


def CDLMORNINGDOJISTAR(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLMORNINGDOJISTAR(popen, phigh, plow, pclose, **kwargs)


def CDLMORNINGSTAR(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLMORNINGSTAR(popen, phigh, plow, pclose, **kwargs)


def CDLONNECK(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLONNECK(popen, phigh, plow, pclose, **kwargs)


def CDLPIERCING(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLPIERCING(popen, phigh, plow, pclose, **kwargs)


def CDLRICKSHAWMAN(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLRICKSHAWMAN(popen, phigh, plow, pclose, **kwargs)


def CDLRISEFALL3METHODS(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLRISEFALL3METHODS(popen, phigh, plow, pclose, **kwargs)


def CDLSEPARATINGLINES(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLSEPARATINGLINES(popen, phigh, plow, pclose, **kwargs)


def CDLSHOOTINGSTAR(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLSHOOTINGSTAR(popen, phigh, plow, pclose, **kwargs)


def CDLSHORTLINE(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLSHORTLINE(popen, phigh, plow, pclose, **kwargs)


def CDLSPINNINGTOP(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLSPINNINGTOP(popen, phigh, plow, pclose, **kwargs)


def CDLSTALLEDPATTERN(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLSTALLEDPATTERN(popen, phigh, plow, pclose, **kwargs)


def CDLSTICKSANDWICH(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLSTICKSANDWICH(popen, phigh, plow, pclose, **kwargs)


def CDLTAKURI(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLTAKURI(popen, phigh, plow, pclose, **kwargs)


def CDLTASUKIGAP(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLTASUKIGAP(popen, phigh, plow, pclose, **kwargs)


def CDLTHRUSTING(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLTHRUSTING(popen, phigh, plow, pclose, **kwargs)


def CDLTRISTAR(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLTRISTAR(popen, phigh, plow, pclose, **kwargs)


def CDLUNIQUE3RIVER(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLUNIQUE3RIVER(popen, phigh, plow, pclose, **kwargs)


def CDLUPSIDEGAP2CROWS(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLUPSIDEGAP2CROWS(popen, phigh, plow, pclose, **kwargs)


def CDLXSIDEGAP3METHODS(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, _ = _extract_ohlc(data)
    return talib.CDLXSIDEGAP3METHODS(popen, phigh, plow, pclose, **kwargs)


# ---------------------------------------------
# Statistic Functions
# ---------------------------------------------
def BETA(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.BETA(phigh, plow, **kwargs)


def CORREL(data, **kwargs):
    _check_talib_presence()
    _, phigh, plow, _, _ = _extract_ohlc(data)
    return talib.CORREL(phigh, plow, **kwargs)


def LINEARREG(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.LINEARREG(prices, **kwargs)


def LINEARREG_ANGLE(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.LINEARREG_ANGLE(prices, **kwargs)


def LINEARREG_INTERCEPT(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.LINEARREG_INTERCEPT(prices, **kwargs)


def LINEARREG_SLOPE(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.LINEARREG_SLOPE(prices, **kwargs)


def STDDEV(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.STDDEV(prices, **kwargs)


def TSF(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.TSF(prices, **kwargs)


def VAR(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.VAR(prices, **kwargs)


# ---------------------------------------------
# Math Transform
# ---------------------------------------------
def ACOS(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.ACOS(prices, **kwargs)


def ASIN(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.ASIN(prices, **kwargs)


def ATAN(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.ATAN(prices, **kwargs)


def CEIL(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.CEIL(prices, **kwargs)


def COS(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.COS(prices, **kwargs)


def COSH(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.COSH(prices, **kwargs)


def EXP(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.EXP(prices, **kwargs)


def FLOOR(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.FLOOR(prices, **kwargs)


def LN(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.LN(prices, **kwargs)


def LOG10(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.LOG10(prices, **kwargs)


def SIN(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.SIN(prices, **kwargs)


def SINH(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.SINH(prices, **kwargs)


def SQRT(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.SQRT(prices, **kwargs)


def TAN(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.TAN(prices, **kwargs)


def TANH(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.TANH(prices, **kwargs)


# ---------------------------------------------
# Math Operators
# ---------------------------------------------
def ADD(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, pvolume = _extract_ohlc(data)
    return talib.ADD(phigh, plow, **kwargs)


def DIV(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, pvolume = _extract_ohlc(data)
    return talib.DIV(phigh, plow, **kwargs)


def MAX(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MAX(prices, **kwargs)


def MAXINDEX(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MAXINDEX(prices, **kwargs)


def MIN(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MIN(prices, **kwargs)


def MININDEX(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MININDEX(prices, **kwargs)


def MINMAX(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MINMAX(prices, **kwargs)


def MINMAXINDEX(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.MINMAXINDEX(prices, **kwargs)


def MULT(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, pvolume = _extract_ohlc(data)
    return talib.MULT(phigh, plow, **kwargs)


def SUB(data, **kwargs):
    _check_talib_presence()
    popen, phigh, plow, pclose, pvolume = _extract_ohlc(data)
    return talib.SUB(phigh, plow, **kwargs)


def SUM(data, **kwargs):
    _check_talib_presence()
    prices = _extract_series(data)
    return talib.SUM(prices, **kwargs)


def SuperTrend(data, **kwargs):
    return supertrend(data.copy(), **kwargs)



from abc import ABCMeta, abstractmethod


class DumbBot:
    _instance = None

    __metaclass__ = ABCMeta

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(DumbBot, cls).__new__(cls)
        return cls._instance

    @abstractmethod
    def start(self, *args, **kwargs):
        pass

    @abstractmethod
    def stop(self, *args, **kwargs):
        pass

    @abstractmethod
    def send_message(self, *args, **kwargs):
        pass

    @abstractmethod
    def send_order(self, order, signal, callback, commands, **kwargs):
        callback(**kwargs)

    @abstractmethod
    def add_connected_listener(self, *args, **kwargs):
        pass

    @abstractmethod
    def add_command_handler(self, *args, **kwargs):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Modified version from QTPyLib: Quantitative Trading Python Library
# https://github.com/ranaroussi/qtpylib
# Copyright 2016-2018 Ran Aroussi
#
# Modified by vin8tech
# Copyright 2019-2021 vin8tech, vinay patil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import atexit
import hashlib
import logging
import os
import time
import sys

# from decimal import *
import decimal

from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from mongoengine import connect

from kinetick.models import Contract
from kinetick.blotter import load_blotter_args, Blotter
from kinetick.enums import SecurityType, PositionType
from kinetick.instrument import Instrument
from kinetick.lib.brokers import Webull, Zerodha
from kinetick.utils import utils
from kinetick.utils.utils import datetime_to_timezone

decimal.getcontext().prec = 5

# =============================================
# check min, python version
if sys.version_info < (3, 4):
    raise SystemError("Kinetick requires Python version >= 3.4")

# =============================================
utils.create_logger(__name__, level=os.getenv('LOGLEVEL') or logging.INFO)


# =============================================


class Broker():
    """Broker class initializer (abstracted, parent class of ``Algo``)
    Current implementation is in hybrid form i.e, streaming data is provided by webull
    and orders are executed on the zerodha. WIP: remove webull dependency.

    :Parameters:

        instruments : list
            List of contract tuples
        zerodha_user : zerodha user id. In next version this will be abstracted to handle multiple brokers.
        zerodha_password: zerodha password
        zerodha_pin: zerodha pin
    """

    __metaclass__ = ABCMeta

    def __init__(self, instruments, zerodha_user=None, zerodha_password=None, zerodha_pin=None, **kwargs):

        # detect running strategy
        self.strategy = str(self.__class__).split('.')[-1].split("'")[0]

        # initialize class logger
        self.log_broker = logging.getLogger(__name__)

        # -----------------------------------
        # assign default vals if not propagated from algo
        if not hasattr(self, 'timezone'):
            self.timezone = "UTC"
        if not hasattr(self, 'tick_window'):
            self.tick_window = 1000
        if not hasattr(self, 'bar_window'):
            self.bar_window = 100
        if not hasattr(self, 'last_price'):
            self.last_price = {}
        if not hasattr(self, 'backtest'):
            self.backtest = False
        if not hasattr(self, 'sms_numbers'):
            self.sms_numbers = []
        if not hasattr(self, 'trade_log_dir'):
            self.trade_log_dir = None
        if not hasattr(self, 'blotter_name'):
            self.blotter_name = None

        self.broker = Webull()
        self.broker.callbacks = self._callback
        self.client = Zerodha(zerodha_user, zerodha_password, zerodha_pin,
                              debug=True) if not self.backtest else None
        self.connect()

        self.log_broker.info("Connection established...")

        # -----------------------------------
        # create contracts
        self.instruments = {}
        self.symbols = []

        self.add_instruments(instruments)

        self.instrument_combos = {}

        self.instrument_obj = {}

        # -----------------------------------
        # track orders & trades
        self.active_trades = {}
        self.trades = []

        # shortcut
        self.account = self.broker.account

        # use: self.orders.pending...
        self.orders = utils.make_object(
            by_tickerid=self.broker.orders,
            by_symbol=self.broker.symbol_orders,
            pending_ttls={},
            pending={},
            filled={},
            active={},
            history={},
            nextId=1,
            recent={}
        )

        # -----------------------------------
        self.db_connection = None

        # -----------------------------------
        # load blotter settings
        self.blotter_args = load_blotter_args(
            self.blotter_name, logger=self.log_broker)
        self.blotter = Blotter(**self.blotter_args)
        self.blotter.args.update(self.blotter.load_cli_args())
        self.blotter_args = self.blotter.args.copy()

        # connect to DB using blotter's settings
        if not self.blotter_args['dbskip']:
            self.mongo_connect()
        # -----------------------------------
        # do stuff on exit
        atexit.register(self._on_exit)

    # ---------------------------------------
    def mongo_connect(self):
        # skip db connection
        if self.blotter_args['dbskip']:
            return

        # already connected?
        if self.db_connection is not None:
            return

        # connect
        params = {
            'host': str(self.blotter_args['dbhost']) if str(self.blotter_args['dbhost']) is not None else 'localhost',
            'port': int(self.blotter_args['dbport']) if self.blotter_args['dbport'] is not None else 27017,
            'username': str(self.blotter_args['dbuser']) if str(self.blotter_args['dbuser']) is not None else None,
            'password': str(self.blotter_args['dbpass']) if str(self.blotter_args['dbpass']) is not None else None,
            'db': str(self.blotter_args['dbname']) if str(self.blotter_args['dbname']) else 'kinetick'
        }
        self.db_connection = connect(**params)

    # ---------------------------------------
    def add_instruments(self, instruments):
        instrument_tuples_dict = {}
        for instrument in instruments:
            try:
                if isinstance(instrument, Contract):
                    instrument = self.broker.contract_to_tuple(instrument)
                else:
                    instrument = utils.create_contract_tuple(instrument)
                contractString = self.broker.contractString(instrument)
                instrument_tuples_dict[contractString] = instrument
                self.broker.createContract(instrument)
            except Exception as e:
                self.log_broker.error("Error creating contract", e)

        self.instruments.update(instrument_tuples_dict)
        self.symbols = list(self.instruments.keys())

    # ---------------------------------------

    @abstractmethod
    def on_fill(self, instrument, order):
        pass

    # -------------------------------------------
    def _on_exit(self):
        self.log_broker.info("Algo stopped...")

        if self.broker is not None:
            self.log_broker.info("Disconnecting...")
            self.broker.disconnect()

        self.log_broker.info("Disconnecting from DB...")
        try:
            self.db_connection.close()
        except Exception as e:
            pass

    # ---------------------------------------
    def connect(self):
        self.broker.connect(stream=False)
        if not self.backtest and self.client:
            self.client.login()

    # ---------------------------------------
    # @abstractmethod
    def _callback(self, caller, msg, **kwargs):

        if caller == "handleHistoricalData":
            # transmit "as-is" to blotter for handling
            self.blotter.callbacks("handleHistoricalData", msg, **kwargs)

        if caller == "handleConnectionClosed":
            self.log_broker.info("Lost connection to Broker...")

            while not self.broker.connected:
                self.connect()
                time.sleep(1.3)
                if not self.broker.connected:
                    print('*', end="", flush=True)

            self.log_broker.info("Connection established...")

        elif caller == "handleOrders":
            if not hasattr(self, "orders"):
                return

            # order canceled? do some cleanup
            if hasattr(msg, 'status') and "CANCELLED" in msg.status.upper():
                if msg.orderId in self.orders.recent.keys():
                    symbol = self.orders.recent[msg.orderId]['symbol']
                    try:
                        del self.orders.pending_ttls[msg.orderId]
                    except Exception as e:
                        pass
                    try:
                        del self.orders.recent[msg.orderId]
                    except Exception as e:
                        pass
                    try:
                        if self.orders.pending[symbol]['orderId'] == msg.orderId:
                            del self.orders.pending[symbol]
                    except Exception as e:
                        pass
                return

            # continue...

            order = self.client.orders[msg.orderId]

            # print("***********************\n\n", order, "\n\n***********************")
            orderId = msg.orderId
            symbol = order["symbol"]

            try:
                try:
                    quantity = self.orders.history[symbol][orderId]['quantity']
                except Exception as e:
                    quantity = self.orders.history[symbol][order['parentId']]['quantity']
                    # ^^ for child orders auto-created by ezibpy
            except Exception as e:
                quantity = 1

            # update pending order to the time actually submitted
            if order["status"] in ["OPENED", "SUBMITTED"]:
                if orderId in self.orders.pending_ttls:
                    self._update_pending_order(symbol, orderId,
                                               self.orders.pending_ttls[orderId],
                                               quantity)

            elif order["status"] == "FILLED":
                self._update_order_history(
                    symbol, orderId, quantity, filled=True)
                self._expire_pending_order(symbol, orderId)
                self._cancel_orphan_orders(orderId)
                self._register_trade(order)

                # filled
                time.sleep(0.005)
                self.on_fill(self.get_instrument(order['symbol']), order)

    # ---------------------------------------
    def _register_trade(self, order):
        """ constructs trade info from order data """
        if order['id'] in self.orders.recent:
            orderId = order['id']
        else:
            orderId = order['parentId']
        # entry / exit?
        symbol = order["symbol"]
        order_data = self.orders.recent[orderId]
        position = self.get_positions(symbol)['position']

        if position != 0:
            # entry
            order_data['action'] = "ENTRY"
            order_data['position'] = position
            order_data['entry_time'] = datetime_to_timezone(
                order['time'])
            order_data['exit_time'] = None
            order_data['entry_order'] = order_data['order_type']
            order_data['entry_price'] = order['avgFillPrice']
            order_data['exit_price'] = 0
            order_data['exit_reason'] = None

        else:
            order_data['action'] = "EXIT"
            order_data['position'] = 0
            order_data['exit_time'] = datetime_to_timezone(order['time'])
            order_data['exit_price'] = order['avgFillPrice']

            # target / stop?
            if order['id'] == order_data['targetOrderId']:
                order_data['exit_reason'] = "TARGET"
            elif order['id'] == order_data['stopOrderId']:
                order_data['exit_reason'] = "STOP"
            else:
                order_data['exit_reason'] = "SIGNAL"

            # remove from collection
            del self.orders.recent[orderId]

        if order_data is None:
            return None

        # trade identifier
        tradeId = self.strategy.upper() + '_' + symbol.upper()
        tradeId = hashlib.sha1(tradeId.encode()).hexdigest()

        # existing trade?
        if tradeId not in self.active_trades:
            self.active_trades[tradeId] = {
                "strategy": self.strategy,
                "action": order_data['action'],
                "quantity": abs(order_data['position']),
                "position": order_data['position'],
                "symbol": order_data["symbol"].split('_')[0],
                "direction": order_data['direction'],
                "entry_time": None,
                "exit_time": None,
                "duration": "0s",
                "exit_reason": order_data['exit_reason'],
                "order_type": order_data['order_type'],
                "market_price": order_data['price'],
                "target": order_data['target'],
                "stop": order_data['initial_stop'],
                "entry_price": 0,
                "exit_price": order_data['exit_price'],
                "realized_pnl": 0
            }
            if "entry_time" in order_data:
                self.active_trades[tradeId]["entry_time"] = order_data['entry_time']
            if "entry_price" in order_data:
                self.active_trades[tradeId]["entry_price"] = order_data['entry_price']
        else:
            # self.active_trades[tradeId]['direction']   = order_data['direction']
            self.active_trades[tradeId]['action'] = order_data['action']
            self.active_trades[tradeId]['position'] = order_data['position']
            self.active_trades[tradeId]['exit_price'] = order_data['exit_price']
            self.active_trades[tradeId]['exit_reason'] = order_data['exit_reason']
            self.active_trades[tradeId]['exit_time'] = order_data['exit_time']

            # calculate trade duration
            try:
                delta = int((self.active_trades[tradeId]['exit_time'] -
                             self.active_trades[tradeId]['entry_time']).total_seconds())
                days, remainder = divmod(delta, 86400)
                hours, remainder = divmod(remainder, 3600)
                minutes, seconds = divmod(remainder, 60)
                duration = ('%sd %sh %sm %ss' %
                            (days, hours, minutes, seconds))
                self.active_trades[tradeId]['duration'] = duration.replace(
                    "0d ", "").replace("0h ", "").replace("0m ", "")
            except Exception as e:
                pass

            trade = self.active_trades[tradeId]
            if trade['entry_price'] > 0 and trade['position'] == 0:
                if trade['direction'] == "SELL":
                    pnl = trade['entry_price'] - trade['exit_price']
                else:
                    pnl = trade['exit_price'] - trade['entry_price']

                pnl = utils.to_decimal(pnl)
                # print("1)", pnl)
                self.active_trades[tradeId]['realized_pnl'] = pnl

        # print("\n\n-----------------")
        # print(self.active_trades[tradeId])
        # print("-----------------\n\n")

        # get trade
        trade = self.active_trades[tradeId].copy()

        # rename trade direction
        trade['direction'] = trade['direction'].replace(
            "BUY", "LONG").replace("SELL", "SHORT")

        # log
        self.log_trade(trade)

        # remove from active trades and add to trade
        if trade['action'] == "EXIT":
            del self.active_trades[tradeId]
            self.trades.append(trade)

        # return trade
        return trade

    # ---------------------------------------
    def log_trade(self, trade):

        # first trade is an exit?
        if trade['entry_time'] is None:
            return

        # connection established
        if self.db_connection is not None:
            pass
            # TODO save to db

        if self.trade_log_dir:
            self.trade_log_dir = (self.trade_log_dir + '/').replace('//', '/')
            trade_log_path = self.trade_log_dir + self.strategy.lower() + "_" + \
                             datetime.now().strftime('%Y%m%d') + ".csv"

            # convert None to empty string !!
            trade.update((k, '') for k, v in trade.items() if v is None)

            # create df
            trade_df = pd.DataFrame(index=[0], data=trade)[[
                'strategy', 'symbol', 'direction', 'quantity', 'entry_time',
                'exit_time', 'exit_reason', 'order_type', 'market_price', 'target',
                'stop', 'entry_price', 'exit_price', 'realized_pnl'
            ]]

            if os.path.exists(trade_log_path):
                trades = pd.read_csv(trade_log_path, header=0)
                trades = trades.append(trade_df, ignore_index=True, sort=True)
                trades.drop_duplicates(['entry_time', 'symbol', 'strategy'],
                                       keep="last", inplace=True)
                trades.to_csv(trade_log_path, header=True, index=False)
                utils.chmod(trade_log_path)
            else:
                trade_df.to_csv(trade_log_path, header=True, index=False)
                utils.chmod(trade_log_path)

    # ---------------------------------------
    def active_order(self, symbol, order_type="STOP"):
        if symbol in self.orders.history:
            for orderId in self.orders.history[symbol]:
                order = self.orders.history[symbol][orderId]
                if order['order_type'].upper() == order_type.upper():
                    return order
        return None

    # ---------------------------------------
    @staticmethod
    def _get_locals(local_params):
        del local_params['self']
        return local_params

    # ---------------------------------------
    def _create_order(self, symbol, direction, quantity, order_type="",
                      limit_price=0, expiry=0, order_id=0, target=0,
                      initial_stop=0, trail_stop_at=0, trail_stop_by=0,
                      stop_limit=False, trail_stop_type='percent', pos_type=None, tradingsymbol=None,
                      sec_type=None, **kwargs):

        # fix prices to comply with contract's min-tick
        ticksize = self.get_contract_details(symbol)['m_minTick']
        limit_price = utils.round_to_fraction(limit_price, ticksize)
        target = utils.round_to_fraction(target, ticksize)
        initial_stop = utils.round_to_fraction(initial_stop, ticksize)
        trail_stop_at = utils.round_to_fraction(trail_stop_at, ticksize)
        trail_stop_by = utils.round_to_fraction(trail_stop_by, ticksize)
        trail_stop_type = "amount" if trail_stop_type == "amount" else "percent"

        self.log_broker.debug('CREATE ORDER: %s %4d %s %s', direction,
                              quantity, symbol, dict(locals(), **kwargs))

        # force BUY/SELL (not LONG/SHORT)
        direction = direction.replace("LONG", "BUY").replace("SHORT", "SELL")

        # modify order?
        if order_type.upper() == "MODIFY":
            self.modify_order(symbol, order_id, quantity, limit_price)
            return

        # continue...

        if "stoploss" in kwargs and initial_stop == 0:
            initial_stop = kwargs['stoploss']

        order_type = Zerodha.ORDER_TYPE_MARKET if limit_price == 0 else Zerodha.ORDER_TYPE_LIMIT
        fillorkill = kwargs["fillorkill"] if "fillorkill" in kwargs else False
        iceberg = kwargs["iceberg"] if "iceberg" in kwargs else False
        tif = kwargs["tif"] if "tif" in kwargs else "DAY"

        # clear expired pending orders
        # self._cancel_expired_pending_orders()

        # # don't submit order if a pending one is waiting
        # if symbol in self.orders.pending:
        #     self.log_broker.warning(
        #         'Not submitting %s order, orders pending: %s', symbol,
        #         self.orders.pending)
        #     return

        # continue...
        order_quantity = abs(quantity)
        if direction.upper() == "SELL":
            order_quantity = -order_quantity

        contract = self.get_contract(symbol)

        # is bracket order
        bracket = (target > 0) | (initial_stop > 0) | (
                trail_stop_at > 0) | (trail_stop_by > 0)
        trigger_price = kwargs['trigger_price'] if 'trigger_price' in kwargs else initial_stop
        # create & submit order
        sec_type = sec_type if sec_type is not None else SecurityType.STOCK
        pos_type = pos_type or (PositionType.CO if bracket else None)

        variety, product = self.client.get_order_variety(sec_type, pos_type)
        # if it's stock trade then use symbol as tradingsymbol else caller should provide tradingsymbol
        tradingsymbol = symbol if sec_type == SecurityType.STOCK else tradingsymbol
        if tradingsymbol is None:
            raise Exception("No tradingsymbol provided")
        exchange = Zerodha.EXCHANGE_NSE if sec_type == SecurityType.STOCK else Zerodha.EXCHANGE_NFO

        order_id = self.client.place_order(variety=variety, tradingsymbol=tradingsymbol,
                                           transaction_type=direction.upper(), quantity=abs(order_quantity),
                                           product=product, order_type=order_type,
                                           exchange=exchange, price=limit_price,
                                           trigger_price=trigger_price, stoploss=initial_stop)

        self.log_broker.debug('PLACED ORDER: %s %s', self.broker.contract_to_dict(
            contract), order_id)

        # triggered trailing stop?
        # if trail_stop_by != 0 and trail_stop_at != 0:
        #     trail_stop_params = {
        #         "symbol": symbol,
        #         "quantity": -order_quantity,
        #         "triggerPrice": trail_stop_at,
        #         "parentId": order["entryOrderId"],
        #         "stopOrderId": order["stopOrderId"]
        #     }
        #     if trail_stop_type.lower() == 'amount':
        #         trail_stop_params["trailAmount"] = trail_stop_by
        #     else:
        #         trail_stop_params["trailPercent"] = trail_stop_by
        #     self.ibConn.createTriggerableTrailingStop(**trail_stop_params)

        # add all orders to history
        self._update_order_history(symbol=symbol,
                                   orderId=order_id,
                                   quantity=order_quantity,
                                   order_type='ENTRY')

        # self._update_order_history(symbol=symbol,
        #                            orderId=order["targetOrderId"],
        #                            quantity=-order_quantity,
        #                            order_type='TARGET',
        #                            parentId=order["entryOrderId"])
        #
        # self._update_order_history(symbol=symbol,
        #                            orderId=order["stopOrderId"],
        #                            quantity=-order_quantity,
        #                            order_type='STOP',
        #                            parentId=order["entryOrderId"])

        # have original params available for FILL event
        self.orders.recent[order_id] = self._get_locals(locals())
        self.orders.recent[order_id]['targetOrderId'] = 0
        self.orders.recent[order_id]['stopOrderId'] = 0

        if bracket:
            pass
            # self.orders.recent[orderId]['targetOrderId'] = order["targetOrderId"]
            # self.orders.recent[orderId]['stopOrderId'] = order["stopOrderId"]

        # append market price at the time of order
        try:
            self.orders.recent[order_id]['price'] = self.last_price[symbol]
        except Exception as e:
            self.orders.recent[order_id]['price'] = 0

        # add orderId / ttl to (auto-adds to history)
        expiry = expiry * 1000 if expiry > 0 else 60000  # 1min
        self._update_pending_order(symbol, order_id, expiry, order_quantity)
        time.sleep(0.1)
        return order_id

    # ---------------------------------------
    def _cancel_order(self, order_id):
        if order_id is not None:
            return self.client.exit_order(order_id)

    # ---------------------------------------
    def modify_order_group(self, symbol, orderId, entry=None,
                           target=None, stop=None, quantity=None):
        """ IMPORTANT! Not yet supported TODO """
        raise Exception("Not supported")
        # order_group = self.orders.recent[orderId]['order']
        #
        # if entry is not None:
        #     self.modify_order(
        #         symbol, orderId, limit_price=entry, quantity=quantity)
        #
        # if target is not None:
        #     self.modify_order(symbol, order_group['targetOrderId'],
        #                       limit_price=target, quantity=quantity)
        # if stop is not None:
        #     stop_quantity = quantity * -1 if quantity is not None else None
        #     self.modify_order(symbol, order_group['stopOrderId'],
        #                       limit_price=stop, quantity=stop_quantity)

    # ---------------------------------------
    def modify_order(self, symbol, orderId, quantity=None, limit_price=None):
        """ IMPORTANT! Not supported yet """
        raise Exception("Not supported")

    # ---------------------------------------
    @staticmethod
    def _milliseconds_delta(delta):
        return delta.days * 86400000 + delta.seconds * 1000 + delta.microseconds / 1000

    # ---------------------------------------
    def _cancel_orphan_orders(self, orderId):
        """ cancel child orders when parent is gone """
        orders = self.client.orders
        for order in orders:
            order = orders[order]
            if order['parentId'] != orderId:
                self.broker.cancelOrder(order['id'])

    # ---------------------------------------
    def _cancel_expired_pending_orders(self):
        """ expires pending orders TODO"""
        # use a copy to prevent errors
        pending = self.orders.pending.copy()
        for symbol in pending:
            orderId = pending[symbol]["orderId"]
            expiration = pending[symbol]["expires"]

            delta = expiration - datetime.now()
            delta = self._milliseconds_delta(delta)

            # cancel order if expired
            if delta < 0:
                self.client.cancel_order(orderId)
                if orderId in self.orders.pending_ttls:
                    if orderId in self.orders.pending_ttls:
                        del self.orders.pending_ttls[orderId]
                    if symbol in self.orders.pending:
                        if self.orders.pending[symbol]['orderId'] == orderId:
                            del self.orders.pending[symbol]

    # ---------------------------------------------------------
    def _expire_pending_order(self, symbol, orderId):
        self.client.cancel_order(order_id=orderId)  # TODO

        if orderId in self.orders.pending_ttls:
            del self.orders.pending_ttls[orderId]

        if symbol in self.orders.pending:
            if self.orders.pending[symbol]['orderId'] == orderId:
                del self.orders.pending[symbol]

    # ---------------------------------------------------------
    def _update_pending_order(self, symbol, orderId, expiry, quantity):
        self.orders.pending[symbol] = {
            "orderId": orderId,
            "quantity": quantity,
            # "created": datetime.now(),
            "expires": datetime.now() + timedelta(milliseconds=expiry)
        }

        # ibCallback needs this to update with submittion time
        self.orders.pending_ttls[orderId] = expiry
        self._update_order_history(
            symbol=symbol, orderId=orderId, quantity=quantity)

    # ---------------------------------------------------------
    def _update_order_history(self, symbol, orderId, quantity,
                              order_type='entry', filled=False, parentId=0):
        if symbol not in self.orders.history:
            self.orders.history[symbol] = {}

        self.orders.history[symbol][orderId] = {
            "orderId": orderId,
            "quantity": quantity,
            "order_type": order_type.upper(),
            "filled": filled,
            "parentId": parentId
        }

    # ---------------------------------------
    # UTILITY FUNCTIONS
    # ---------------------------------------
    def get_instrument(self, symbol):
        """
        A string subclass that provides easy access to misc
        symbol-related methods and information using shorthand.
        Refer to the `Instruments API <#instrument-api>`_
        for available methods and properties

        Call from within your strategy:
        ``instrument = self.get_instrument("SYMBOL")``

        :Parameters:
            symbol : string
                instrument symbol

        """
        symbol = self.get_symbol(symbol)
        if symbol in self.instrument_obj:
            return self.instrument_obj[symbol]
        else:
            instrument = Instrument(symbol)
            instrument._set_parent(self)
            instrument._set_windows(ticks=self.tick_window, bars=self.bar_window)
            self.instrument_obj[symbol] = instrument
            return instrument

    # ---------------------------------------
    @staticmethod
    def get_symbol(symbol):
        if not isinstance(symbol, str):
            if isinstance(symbol, dict):
                symbol = symbol['symbol']
            elif isinstance(symbol, pd.DataFrame):
                symbol = symbol[:1]['symbol'].values[0]

        return symbol

    # ---------------------------------------
    def get_account(self):
        return self.broker.account

    # ---------------------------------------
    def get_contract(self, symbol):
        return self.broker.contracts[self.broker.tickerId(symbol)]

    # ---------------------------------------
    def get_contract_details(self, symbol):
        return self.broker.contractDetails(symbol)

    # ---------------------------------------
    def get_tickerId(self, symbol):
        return self.broker.tickerId(symbol)

    # ---------------------------------------
    def get_orders(self, symbol):
        symbol = self.get_symbol(symbol)

        self.orders.by_symbol = self.broker.group_orders("symbol")
        if symbol in self.orders.by_symbol:
            return self.orders.by_symbol[symbol]

        return {}

    # ---------------------------------------
    def get_positions(self, symbol):
        symbol = self.get_symbol(symbol)

        if self.backtest:
            position = 0
            avgCost = 0.0

            if self.datastore.recorded is not None:
                data = self.datastore.recorded
                col = symbol.upper() + '_POSITION'
                position = data[col].values[-1]
                if position != 0:
                    pos = data[col].diff()
                    avgCost = data[data.index.isin(pos[pos != 0][-1:].index)
                    ][symbol.upper() + '_OPEN'].values[-1]
            return {
                "symbol": symbol,
                "position": position,
                "avgCost": avgCost,
                "account": "Backtest"
            }

        elif symbol in self.broker.positions:
            return self.broker.positions[symbol]

        return {
            "symbol": symbol,
            "position": 0,
            "avgCost": 0.0,
            "account": None
        }

    # ---------------------------------------
    def get_portfolio(self, symbol=None):
        raise Exception("Not supported")
        # if symbol is not None:
        #     symbol = self.get_symbol(symbol)
        #
        #     if symbol in self.zerodha.portfolio:
        #         portfolio = self.zerodha.portfolio[symbol]
        #         if "symbol" in portfolio:
        #             return portfolio
        #
        #     return {
        #         "symbol": symbol,
        #         "position": 0.0,
        #         "marketPrice": 0.0,
        #         "marketValue": 0.0,
        #         "averageCost": 0.0,
        #         "unrealizedPNL": 0.0,
        #         "realizedPNL": 0.0,
        #         "totalPNL": 0.0,
        #         "account": None
        #     }
        #
        # return self.zerodha.portfolio

    # ---------------------------------------
    def get_pending_orders(self, symbol=None):
        if symbol is not None:
            symbol = self.get_symbol(symbol)
            if symbol in self.orders.pending:
                return self.orders.pending[symbol]
            return {}

        return self.orders.pending

    # ---------------------------------------
    def get_trades(self, symbol=None):

        # closed trades
        trades = pd.DataFrame(self.trades)
        if not trades.empty:
            trades.loc[:, 'closed'] = True

        # ongoing trades
        active_trades = pd.DataFrame(list(self.active_trades.values()))
        if not active_trades.empty:
            active_trades.loc[:, 'closed'] = False

        # combine dataframes
        df = pd.concat([trades, active_trades], sort=True).reset_index()

        # set last price
        if not df.empty:

            # conert values to floats
            df['entry_price'] = df['entry_price'].astype(float)
            df['exit_price'] = df['exit_price'].astype(float)
            df['market_price'] = df['market_price'].astype(float)
            df['realized_pnl'] = df['realized_pnl'].astype(float)
            df['stop'] = df['stop'].astype(float)
            df['target'] = df['target'].astype(float)
            df['quantity'] = df['quantity'].astype(int)

            try:
                df.loc[:, 'last'] = self.last_price[symbol]
            except Exception as e:
                df.loc[:, 'last'] = 0

            # calc unrealized pnl
            df['unrealized_pnl'] = np.where(df['direction'] == "SHORT",
                                            df['entry_price'] - df['last'],
                                            df['last'] - df['entry_price'])

            df.loc[df['closed'], 'unrealized_pnl'] = 0

            # drop index column
            df.drop('index', axis=1, inplace=True)

            # get single symbol
            if symbol is not None:
                df = df[df['symbol'] == symbol.split("_")[0]]
                df.loc[:, 'symbol'] = symbol

        # return
        return df



from mongoengine import FloatField, StringField, DateTimeField, IntField, DynamicDocument
import datetime

from kinetick.enums import Timeframes

RESOLUTION = Timeframes.tuple()


class OHLC(DynamicDocument):
    tickerId = StringField(max_length=50, required=True)
    symbol = StringField(max_length=50, required=True)
    datetime = DateTimeField(required=True, default=datetime.datetime.utcnow)
    open = FloatField(required=True)
    high = FloatField(required=True)
    low = FloatField(required=True)
    close = FloatField(required=True)
    volume = IntField(required=True)
    vwap = FloatField()
    resolution = StringField(choices=RESOLUTION, default=Timeframes.MINUTE_T, unique_with=['tickerId', 'datetime'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Modified version from QTPyLib: Quantitative Trading Python Library
# https://github.com/ranaroussi/qtpylib
# Copyright 2016-2018 Ran Aroussi
#
# Modified by vin8tech
# Copyright 2019-2021 vin8tech, vinay patil
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import inspect
import sys
import logging
import os

from datetime import datetime
from abc import ABCMeta, abstractmethod

import warnings

import pandas as pd
from kinetick.models import Position
from numpy import nan

from kinetick.utils import utils, asynctools
from kinetick.broker import Broker
from kinetick.enums import Timeframes, PositionType
from kinetick.risk_assessor import RiskAssessor
from kinetick.utils.utils import create_logger
from kinetick.workflow import validate_columns as validate_csv_columns
from kinetick.blotter import prepare_history

# =============================================
warnings.simplefilter(action='ignore', category=FutureWarning)

# check min, python version
if sys.version_info < (3, 4):
    raise SystemError("Kinetick requires Python version >= 3.4")

# =============================================
# Configure logging
create_logger(__name__, level=os.getenv('LOGLEVEL') or logging.INFO)

# =============================================
# set up threading pool
__threads__ = utils.read_single_argv("--threads")
__threads__ = int(__threads__) if utils.is_number(__threads__) else None
asynctools.multitasking.createPool(__name__, __threads__)


# =============================================


class Algo(Broker):
    """Algo class initializer (sub-class of Broker)

    :Parameters:

        instruments : list
            List of contract tuples. Default is empty list
        resolution : str
            Desired bar resolution (using pandas resolution: 1T, 1H, etc).
            Use K for tick bars. Default is 1T (1min)
        tick_window : int
            Length of tick lookback window to keep. Defaults to 1
        bar_window : int
            Length of bar lookback window to keep. Defaults to 100
        timezone : str
            Convert broker timestamps to this timezone.
            Defaults to UTC
        preload : str
            Preload history when starting algo (Pandas resolution: 1H, 1D, etc)
            Use K for tick bars.
        continuous : bool
            Tells preloader to construct continuous Futures contracts
            (default is True)
        blotter : str
            Log trades to this Blotter's Datastore (default is "auto detect")
        log: str
            Path to store trade data (default: None)
        backtest: bool
            Whether to operate in Backtest mode (default: False)
        start: str
            Backtest start date (YYYY-MM-DD [HH:MM:SS[.MS]). Default is None
        end: str
            Backtest end date (YYYY-MM-DD [HH:MM:SS[.MS]). Default is None
        data : str
            Path to the directory with Kinetick-compatible CSV files (Backtest)
        output: str
            Path to save the recorded data (default: None)
        risk_assessor: RiskAssessor
            All instances of Algo will use same global risk assessor if provided.
        name: String
            Strategy name. Please provide unique identifier to your algo if running multiple algos.
    """

    __metaclass__ = ABCMeta

    def __init__(self, instruments, risk_assessor: RiskAssessor = None, resolution="1m",
                 tick_window=1, bar_window=100, timezone="UTC", preload=None,
                 continuous=True, blotter=None, sms=None, log=None,
                 backtest=False, start=None, end=None, data=None, output=None,
                 backfill=False, name=None, preload_positions=None, **kwargs):

        # detect algo name
        self.name = name or str(self.__class__).split('.')[-1].split("'")[0]

        # initialize algo logger
        self.log_algo = logging.getLogger(__name__)

        # initialize strategy logger
        utils.create_logger(self.name, level=logging.INFO)
        self.log = logging.getLogger(self.name)

        # override args with any (non-default) command-line args
        self.args = {arg: val for arg, val in locals().items(
        ) if arg not in ('__class__', 'self', 'kwargs')}
        self.args.update(kwargs)
        self.args.update(self.load_cli_args())

        # -----------------------------------
        # assign algo params
        self.bars = pd.DataFrame(columns=["symbol", "symbol_group"])
        self.ticks = pd.DataFrame(columns=["symbol", "symbol_group"])
        self.quotes = {}
        self.books = {}
        self.tick_count = 0
        self.tick_bar_count = 0
        self.bar_count = 0
        self.bar_hashes = {}

        self.tick_window = tick_window if tick_window > 0 else 1
        if "V" in resolution:
            self.tick_window = 1000
        self.bar_window = bar_window if bar_window > 0 else 100
        self.resolution = Timeframes.timeframe_to_resolution(resolution)
        self.timezone = timezone
        self.preload = preload
        self.preload_positions = self.args['preload_positions']
        self.backfill = self.args["backfill"]
        self.continuous = continuous

        # -----------------------------------
        # backtest info
        self.backtest = self.args["backtest"]
        self.backtest_start = self.args["start"]
        self.backtest_end = self.args["end"]
        self.backtest_csv = self.args["data"]

        # -----------------------------------
        self.sms_numbers = self.args["sms"]
        self.trade_log_dir = self.args["log"]
        self.blotter_name = self.args["blotter"]
        self.record_output = self.args["output"]

        self.risk_assessor = risk_assessor if risk_assessor is not None else RiskAssessor(**self.args)

        # ---------------------------------------
        # sanity checks for backtesting mode
        if self.backtest:
            if self.record_output is None:
                self.log_algo.error(
                    "Must provide an output file for Backtest mode")
                sys.exit(0)
            if self.backtest_start is None:
                self.log_algo.error(
                    "Must provide start date for Backtest mode")
                sys.exit(0)
            if self.backtest_end is None:
                self.backtest_end = datetime.now().strftime(
                    '%Y-%m-%d %H:%M:%S.%f')
            if self.backtest_csv is not None:
                if not os.path.exists(self.backtest_csv):
                    self.log_algo.error(
                        "CSV directory cannot be found (%s)",
                        self.backtest_csv)
                    sys.exit(0)
                elif self.backtest_csv.endswith("/"):
                    self.backtest_csv = self.backtest_csv[:-1]

        else:
            self.backtest_start = None
            self.backtest_end = None
            self.backtest_csv = None

        # -----------------------------------
        # initiate broker/order manager
        super().__init__(instruments,
                         zerodha_user=self.args['zerodha_user'],
                         zerodha_password=self.args['zerodha_password'],
                         zerodha_pin=self.args['zerodha_pin'])

        # -----------------------------------
        # signal collector
        self.signals = {}
        for sym in self.symbols:
            self.signals[sym] = pd.DataFrame()

        # -----------------------------------
        # initialize output file
        self.record_ts = None
        if self.record_output:
            self.datastore = utils.DataStore(self.args["output"])

        # ---------------------------------------
        # add stale ticks for more accurate time--based bars
        if not self.backtest and self.resolution[-1] not in ("S", "K", "V"):
            self.bar_timer = asynctools.RecurringTask(
                self.add_stale_tick, interval_sec=1, init_sec=1, daemon=True)

        # ---------------------------------------
        # be aware of thread count
        self.threads = asynctools.multitasking.getPool(__name__)['threads']

    # ---------------------------------------
    def add_stale_tick(self):
        ticks = self.ticks.copy()
        if self.ticks.empty:
            return

        last_tick_sec = float(utils.datetime64_to_datetime(
            ticks.index.values[-1]).strftime('%M.%S'))

        for sym in list(self.ticks["symbol"].unique()):
            tick = ticks[ticks['symbol'] ==
                         sym][-5:].to_dict(orient='records')[-1]
            tick['timestamp'] = datetime.utcnow()

            if last_tick_sec != float(tick['timestamp'].strftime("%M.%S")):
                tick = pd.DataFrame(index=[0], data=tick)
                tick.set_index('timestamp', inplace=True)
                tick = utils.set_timezone(tick, tz=self.timezone)
                tick.loc[:, 'lastsize'] = 0  # no real size

                self._tick_handler(tick, stale_tick=True)

    # ---------------------------------------
    def load_cli_args(self):
        """
        Parse command line arguments and return only the non-default ones

        :Retruns: dict
            a dict of any non-default args passed on the command-line.
        """
        parser = argparse.ArgumentParser(
            description='Kinetick Algo',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument('--sms', default=self.args["sms"],
                            help='Numbers to text orders', nargs='+')
        parser.add_argument('--log', default=self.args["log"],
                            help='Path to store trade data')
        parser.add_argument('--backtest', default=self.args["backtest"],
                            help='Work in Backtest mode (flag)',
                            action='store_true')
        parser.add_argument('--backfill', default=self.args["backfill"],
                            help='backfill bar data (flag)',
                            action='store_true')
        parser.add_argument('--start', default=self.args["start"],
                            help='Backtest start date')
        parser.add_argument('--end', default=self.args["end"],
                            help='Backtest end date')
        parser.add_argument('--data', default=self.args["data"],
                            help='Path to backtester CSV files')
        parser.add_argument('--output', default=self.args["output"],
                            help='Path to save the recorded data')
        parser.add_argument('--blotter',
                            help='Log trades to this Blotter\'s Datastore')
        parser.add_argument('--continuous', default=self.args["continuous"],
                            help='Use continuous Futures contracts (flag)',
                            action='store_true')
        parser.add_argument('--zerodha_user', default=os.getenv("zerodha_user"),
                            help='Zerodha Username', required=False)
        parser.add_argument('--zerodha_password', default=os.getenv("zerodha_password"),
                            help='Zerodha Password', required=False)
        parser.add_argument('--zerodha_pin', default=os.getenv("zerodha_pin"),
                            help='Zerodha PIN', required=False)
        parser.add_argument('--resolution', default=os.getenv("resolution") or Timeframes.MINUTE_1,
                            help='Bar interval in terms of resolution (default=1m). '
                                 'ex. 1m, 3m, 5m, 15m, 30m, 1h, 2h, 3h, 4h, 6h, 8h, 1D',
                            required=False)
        parser.add_argument('--max_trades', default=os.getenv("max_trades") or 1, type=int,
                            help='Max Active Concurrent Trades (default=1). ex. 4', required=False)
        parser.add_argument('--initial_capital', default=os.getenv("initial_capital") or 10000, type=float,
                            help='Initial Capital (default=10000). ex. 1200000', required=False)
        parser.add_argument('--initial_margin', default=os.getenv("initial_margin") or 1000, type=float,
                            help='Initial Margin (default=1000). ex. 10000', required=False)
        parser.add_argument('--risk2reward', default=os.getenv("risk2reward") or 1, type=float,
                            help='Risk to Reward (default=1). ex. 1.2', required=False)
        parser.add_argument('--risk_per_trade', default=os.getenv("risk_per_trade") or 100, type=float,
                            help='Risk per Trade (default=100), ex. 100', required=False)
        parser.add_argument('--preload_positions',
                            default=os.getenv("preload_positions") or self.args["preload_positions"]
                            if 'preload_positions' in self.args else None,
                            help='Preload positions. Only available in live trade. ex. 4D, 1H', required=False)

        cmd_args, _ = parser.parse_known_args()
        args = {arg: val for arg, val in vars(
            cmd_args).items()}
        return args

    # ---------------------------------------
    def run(self):
        """Starts the algo

        Connects to the Blotter, processes market data and passes
        tick data to the ``on_tick`` function and bar data to the
        ``on_bar`` methods.
        """

        history = pd.DataFrame()

        # get history from csv dir
        if self.backtest and self.backtest_csv:
            kind = "TICK" if self.resolution[-1] in ("S", "K", "V") else "BAR"
            dfs = []
            for symbol in self.symbols:
                file = "%s/%s.%s.csv" % (self.backtest_csv, symbol, kind)
                if not os.path.exists(file):
                    self.log_algo.error(
                        "Can't load data for %s (%s doesn't exist)",
                        symbol, file)
                    sys.exit(0)
                try:
                    df = pd.read_csv(file)
                    if "expiry" not in df.columns:
                        df.loc[:, "expiry"] = nan

                    if not validate_csv_columns(df, kind, raise_errors=False):
                        self.log_algo.error(
                            "%s isn't a Kinetick-compatible format", file)
                        sys.exit(0)

                    if df['symbol'].values[-1] != symbol:
                        self.log_algo.error(
                            "%s Doesn't content data for %s", file, symbol)
                        sys.exit(0)

                    dfs.append(df)

                except Exception as e:
                    self.log_algo.error(
                        "Error reading data for %s (%s)", symbol, file)
                    sys.exit(0)

            history = prepare_history(
                data=pd.concat(dfs, sort=True),
                resolution=self.resolution,
                tz=self.timezone,
                continuous=self.continuous
            )
            history = history[history.index >= self.backtest_start]

        elif not self.blotter_args["dbskip"] and (
                self.backtest or self.preload):

            start = self.backtest_start if self.backtest else utils.backdate(
                self.preload)
            end = self.backtest_end if self.backtest else None

            history = self.blotter.history(
                symbols=self.symbols,
                start=start,
                end=end,
                resolution=Timeframes.to_timeframe(self.resolution),
                tz=self.timezone,
                continuous=self.continuous
            )

            # history needs backfilling?
            # self.blotter.backfilled = True
            if not self.blotter.backfilled and self.backfill:
                # "loan" Blotter this stream provider
                self.blotter.connection = self.broker

                # call the back fill
                self.blotter.backfill(data=history,
                                      resolution=Timeframes.to_timeframe(self.resolution),
                                      start=start, end=end)

                # re-get history from db
                history = self.blotter.history(
                    symbols=self.symbols,
                    start=start,
                    end=end,
                    resolution=Timeframes.to_timeframe(self.resolution),
                    tz=self.timezone,
                    continuous=self.continuous
                )

                # take our connection back :)
                self.blotter.connection = None

        # optimize pandas
        if not history.empty:
            history['symbol'] = history['symbol'].astype('category')
            history['symbol_group'] = history['symbol_group'].astype('category')
            history['asset_class'] = history['asset_class'].astype('category')
            history = history.loc[history['volume'] > 0]

        if self.backtest:
            # initiate strategy
            self.on_start()

            # drip history
            drip_handler = self._tick_handler if self.resolution[-1] in (
                "S", "K", "V") else self._bar_handler
            self.blotter.drip(history, drip_handler)

        else:
            if not self.blotter_args["dbskip"] and self.preload_positions:
                start = utils.backdate(self.preload_positions)
                self.load_positions(start)
            # place history self.bars
            self.bars = history

            # add instruments to blotter in case they do not exist
            self.blotter.register(self.instruments)

            # initiate strategy
            self.on_start()

            # listen for RT data
            self.blotter.stream(
                symbols=self.symbols,
                tz=self.timezone,
                quote_handler=self._quote_handler,
                tick_handler=self._tick_handler,
                bar_handler=self._bar_handler,
                book_handler=self._book_handler
            )

    # ---------------------------------------
    @abstractmethod
    def on_start(self):
        """
        Invoked once when algo starts. Used for when the strategy
        needs to initialize parameters upon starting.

        """
        # raise NotImplementedError("Should implement on_start()")
        pass

    # ---------------------------------------
    @abstractmethod
    def on_quote(self, instrument):
        """
        Invoked on every quote captured for the selected instrument.
        This is where you'll write your strategy logic for quote events.

        :Parameters:

            symbol : string
                `Instruments Object <#instrument-api>`_

        """
        # raise NotImplementedError("Should implement on_quote()")
        pass

    # ---------------------------------------
    @abstractmethod
    def on_tick(self, instrument):
        """
        Invoked on every tick captured for the selected instrument.
        This is where you'll write your strategy logic for tick events.

        :Parameters:

            symbol : string
                `Instruments Object <#instrument-api>`_

        """
        # raise NotImplementedError("Should implement on_tick()")
        pass

    # ---------------------------------------
    @abstractmethod
    def on_bar(self, instrument):
        """
        Invoked on every tick captured for the selected instrument.
        This is where you'll write your strategy logic for tick events.

        :Parameters:

            instrument : object
                `Instruments Object <#instrument-api>`_

        """
        # raise NotImplementedError("Should implement on_bar()")
        pass

    # ---------------------------------------
    @abstractmethod
    def on_orderbook(self, instrument):
        """
        Invoked on every change to the orderbook for the selected instrument.
        This is where you'll write your strategy logic for orderbook events.

        :Parameters:

            symbol : string
                `Instruments Object <#instrument-api>`_

        """
        # raise NotImplementedError("Should implement on_orderbook()")
        pass

    # ---------------------------------------
    @abstractmethod
    def on_fill(self, instrument, order):
        """
        Invoked on every order fill for the selected instrument.
        This is where you'll write your strategy logic for fill events.

        :Parameters:

            instrument : object
                `Instruments Object <#instrument-api>`_
            order : object
                Filled order data

        """
        # raise NotImplementedError("Should implement on_fill()")
        pass

    # ---------------------------------------
    def get_history(self, symbols, start, end=None, resolution="1T", tz="UTC"):
        """Get historical market data.
        Connects to Blotter and gets historical data from storage

        :Parameters:
            symbols : list
                List of symbols to fetch history for
            start : datetime / string
                History time period start date
                datetime or YYYY-MM-DD[ HH:MM[:SS]] string)

        :Optional:
            end : datetime / string
                History time period end date
                (datetime or YYYY-MM-DD[ HH:MM[:SS]] string)
            resolution : string
                History resolution (Pandas resample, defaults to 1T/1min)
            tz : string
                History timezone (defaults to UTC)

        :Returns:
            history : pd.DataFrame
                Pandas DataFrame object with historical data for all symbols
        """
        return self.blotter.history(symbols, start, end, resolution, tz)

    # ---------------------------------------
    # shortcuts to broker._create_order
    # ---------------------------------------
    def order(self, txn_type, symbol, quantity=0, **kwargs):
        """ Send an order for the selected instrument

        :Parameters:

            txn_type : string
                Order Type (BUY/SELL, EXIT/FLATTEN)
            symbol : string
                instrument symbol
            quantity : int
                Order quantiry

        :Optional:

            limit_price : float
                In case of a LIMIT order, this is the LIMIT PRICE
            expiry : int
                Cancel this order if not filled after *n* seconds
                (default 60 seconds)
            order_type : string
                Type of order: Market (default),
                LIMIT (default when limit_price is passed),
                MODIFY (required passing or orderId)
            orderId : int
                If modifying an order, the order id of the modified order
            target : float
                Target (exit) price
            initial_stop : float
                Price to set hard stop
            stop_limit: bool
                Flag to indicate if the stop should be STOP or STOP LIMIT.
                Default is ``False`` (STOP)
            trail_stop_at : float
                Price at which to start trailing the stop
            trail_stop_type : string
                Type of traiing stop offset (amount, percent).
                Default is ``percent``
            trail_stop_by : float
                Offset of trailing stop distance from current price
            fillorkill: bool
                Fill entire quantiry or none at all
            iceberg: bool
                Is this an iceberg (hidden) order
            tif: str
                Time in force (DAY, GTC, IOC, GTD). default is ``DAY``
        """
        self.log_algo.info('ORDER: %s %4d %s %s', txn_type,
                           quantity, symbol, kwargs)
        if txn_type.upper() == "EXIT" or txn_type.upper() == "FLATTEN":
            position = self.get_positions(symbol)
            if position['position'] == 0:
                return

            kwargs['symbol'] = symbol
            kwargs['quantity'] = abs(position['position'])
            kwargs['direction'] = "BUY" if position['position'] < 0 else "SELL"

            # print("EXIT", kwargs)

            try:
                self.record({symbol + '_POSITION': 0}, **kwargs)
            except Exception as e:
                pass

            if not self.backtest:
                self._create_order(**kwargs)

        else:
            if quantity == 0:
                return

            # buy/sell at best price
            limit_price = kwargs['limit_price'] if 'limit_price' in kwargs else 0
            best_price = kwargs['best_price'] if 'best_price' in kwargs else False

            if best_price is True and limit_price > 0:
                instrument = self.get_instrument(symbol)
                order_book = instrument.get_orderbook()
                running_price = float(order_book['bid'][0]) if txn_type.upper() != "BUY" else float(
                    order_book['ask'][0])
                running_size = float(order_book['bidsize'][0]) if txn_type.upper() != "BUY" else float(
                    order_book['asksize'][0])

                if running_price > 0 and running_size >= abs(quantity):
                    if txn_type.upper() != "BUY":
                        kwargs['limit_price'] = running_price if running_price > limit_price else limit_price
                    else:
                        kwargs['limit_price'] = running_price if running_price < limit_price else limit_price

            kwargs['symbol'] = symbol
            kwargs['quantity'] = abs(quantity)
            kwargs['direction'] = txn_type.upper()
            # kwargs['trigger_price'] = running_price if running_price > 0 else limit_price

            try:
                quantity = abs(quantity)
                if kwargs['direction'] != "BUY":
                    quantity = -quantity
                self.record({'POSITION': quantity}, **kwargs)
            except Exception as e:
                pass

            if not self.backtest:
                return self._create_order(**kwargs)

    # ---------------------------------------
    def cancel_order(self, order_id):
        """ Cancels a un-filled order

        Parameters:
            order_id : int
                Order ID
        """
        if not self.backtest:
            return self._cancel_order(order_id)

    # ---------------------------------------
    def record(self, *args, **kwargs):
        """Records data for later analysis.
        Values will be logged to the file specified via
        ``--output [file]`` (along with bar data) as
        csv/pickle/h5 file.

        Call from within your strategy:
        ``self.record(key=value)``

        :Parameters:
            ** kwargs : mixed
                The names and values to record

        """
        if self.record_output:
            try:
                self.datastore.record(self.record_ts, *args, **kwargs)
            except Exception as e:
                self.log_algo.error('Error Recording', e)

    # ---------------------------------------
    @staticmethod
    def _caller(caller):
        stack = [x[3] for x in inspect.stack()][1:-1]
        return caller in stack

    # ---------------------------------------
    @asynctools.multitasking.task
    def _book_handler(self, book):
        symbol = book['symbol']
        del book['symbol']
        del book['kind']

        self.books[symbol] = book
        self.on_orderbook(self.get_instrument(symbol))

    # ---------------------------------------
    @asynctools.multitasking.task
    def _quote_handler(self, quote):
        del quote['kind']
        self.quotes[quote['symbol']] = quote
        self.on_quote(self.get_instrument(quote))

    # ---------------------------------------
    @staticmethod
    def _get_window_per_symbol(df, window):
        # truncate tick window per symbol
        dfs = []
        for sym in list(df["symbol"].unique()):
            dfs.append(df[df['symbol'] == sym][-window:])
        return pd.concat(dfs, sort=True).sort_index()

    # ---------------------------------------
    @staticmethod
    def _thread_safe_merge(symbol, basedata, newdata):
        data = newdata
        if "symbol" in basedata.columns:
            data = pd.concat(
                [basedata[basedata['symbol'] != symbol], data], sort=True)

        data.loc[:, '_idx_'] = data.index
        data = data.drop_duplicates(
            subset=['_idx_', 'symbol', 'symbol_group', 'asset_class'],
            keep='last')
        data = data.drop('_idx_', axis=1)
        data = data.sort_index()

        try:
            return data.dropna(subset=[
                'open', 'high', 'low', 'close', 'volume'])
        except Exception as e:
            return data

    # ---------------------------------------
    @asynctools.multitasking.task
    def _tick_handler(self, tick, stale_tick=False):
        # self._cancel_expired_pending_orders() TODO

        # tick symbol
        symbol = tick['symbol'].values
        if len(symbol) == 0:
            return
        symbol = symbol[0]
        self.last_price[symbol] = float(tick['last'].values[0])

        # work on copy
        self_ticks = self.ticks.copy()

        # initial value
        if self.record_ts is None:
            self.record_ts = tick.index[0]

        if self.resolution[-1] not in ("S", "K", "V"):
            if self.threads == 0:
                self.ticks = self._update_window(
                    self.ticks, tick, window=self.tick_window)
            else:
                self_ticks = self._update_window(
                    self_ticks, tick, window=self.tick_window)
                self.ticks = self._thread_safe_merge(
                    symbol, self.ticks, self_ticks)  # assign back
        else:
            self.ticks = self._update_window(self.ticks, tick)
            # bars = utils.resample(self.ticks, self.resolution)
            bars = utils.resample(
                self.ticks, self.resolution, tz=self.timezone)

            if len(bars.index) > self.tick_bar_count > 0 or stale_tick:
                self.record_ts = tick.index[0]
                self._base_bar_handler(bars[bars['symbol'] == symbol][-1:])

                window = int(
                    "".join([s for s in self.resolution if s.isdigit()]))
                if self.threads == 0:
                    self.ticks = self._get_window_per_symbol(
                        self.ticks, window)
                else:
                    self_ticks = self._get_window_per_symbol(
                        self_ticks, window)
                    self.ticks = self._thread_safe_merge(
                        symbol, self.ticks, self_ticks)  # assign back

            self.tick_bar_count = len(bars.index)

            # record non time-based bars
            self.record(bars[-1:])

        if not stale_tick:
            if self.ticks[(self.ticks['symbol'] == symbol) | (
                    self.ticks['symbol_group'] == symbol)].empty:
                return
            tick_instrument = self.get_instrument(tick)
            if tick_instrument:
                self.on_tick(tick_instrument)

    # ---------------------------------------
    def _base_bar_handler(self, bar):
        """ non threaded bar handler (called by threaded _tick_handler) """
        # bar symbol
        symbol = bar['symbol'].values
        if len(symbol) == 0:
            return
        symbol = symbol[0]
        self_bars = self.bars.copy()  # work on copy

        is_tick_or_volume_bar = False
        handle_bar = True

        if self.resolution[-1] in ("S", "K", "V"):
            is_tick_or_volume_bar = True
            handle_bar = self._caller("_tick_handler")

        # drip is also ok
        handle_bar = handle_bar or self._caller("drip")

        if is_tick_or_volume_bar:
            # just add a bar (used by tick bar bandler)
            if self.threads == 0:
                self.bars = self._update_window(self.bars, bar,
                                                window=self.bar_window)
            else:
                self_bars = self._update_window(self_bars, bar,
                                                window=self.bar_window)
        else:
            # add the bar and resample to resolution
            if self.threads == 0:
                self.bars = self._update_window(self.bars, bar,
                                                window=self.bar_window,
                                                resolution=self.resolution)
            else:
                self_bars = self._update_window(self_bars, bar,
                                                window=self.bar_window,
                                                resolution=self.resolution)

        # assign new data to self.bars if threaded
        if self.threads > 0:
            self.bars = self._thread_safe_merge(symbol, self.bars, self_bars)

        # optimize pandas
        if len(self.bars) == 1:
            self.bars['symbol'] = self.bars['symbol'].astype('category')
            self.bars['symbol_group'] = self.bars['symbol_group'].astype('category')
            self.bars['asset_class'] = self.bars['asset_class'].astype('category')

        # new bar?
        hash_string = bar[:1]['symbol'].to_string().translate(
            str.maketrans({key: None for key in "\n -:+"}))
        this_bar_hash = abs(hash(hash_string)) % (10 ** 8)

        newbar = True
        if symbol in self.bar_hashes.keys():
            newbar = self.bar_hashes[symbol] != this_bar_hash
        self.bar_hashes[symbol] = this_bar_hash

        if newbar and handle_bar:
            if self.bars[(self.bars['symbol'] == symbol) | (
                    self.bars['symbol_group'] == symbol)].empty:
                return
            bar_instrument = self.get_instrument(symbol)
            if bar_instrument:
                self.record_ts = bar.index[0]
                self.on_bar(bar_instrument)
                # if self.resolution[-1] not in ("S", "K", "V"):
                # self.record(bar)

    # ---------------------------------------
    @asynctools.multitasking.task
    def _bar_handler(self, bar):
        """ threaded version of _base_bar_handler (called by blotter's) """
        self._base_bar_handler(bar)

    # ---------------------------------------
    def _update_window(self, df, data, window=None, resolution=None):
        if df is None:
            df = data
        else:
            df = df.append(data, sort=True)
            df.loc[:, '_idx_'] = df.index
            df.drop_duplicates(
                subset=['_idx_', 'symbol', 'symbol_group', 'asset_class'],
                keep='last', inplace=True)
            df.drop('_idx_', axis=1, inplace=True)

        # resample
        if resolution:
            tz = str(df.index.tz)
            # try:
            #     tz = str(df.index.tz)
            # except Exception as e:
            #     tz = None
            df = utils.resample(df, resolution=resolution, tz=tz, sync_last_timestamp=False)

        else:
            # remove duplicates rows
            # (handled by resample if resolution is provided)
            df.loc[:, '_idx_'] = df.index
            df.drop_duplicates(
                subset=['_idx_', 'symbol', 'symbol_group', 'asset_class'],
                keep='last', inplace=True)
            df.drop('_idx_', axis=1, inplace=True)

        # return
        if window is None:
            return df

        # return df[-window:]
        return self._get_window_per_symbol(df, window)

    # ---------------------------------------
    # signal logging methods
    # ---------------------------------------
    def _add_signal_history(self, df, symbol):
        """ Initilize signal history """
        if symbol not in self.signals.keys() or len(self.signals[symbol]) == 0:
            self.signals[symbol] = [nan] * len(df.index)
        else:
            self.signals[symbol].append(nan)

        self.signals[symbol] = self.signals[symbol][-len(df.index):]
        signal_count = len(self.signals[symbol])
        df.loc[-signal_count:, 'signal'] = self.signals[symbol][-signal_count:]

        return df

    def _log_signal(self, symbol, signal):
        """ Log signal

        :Parameters:
            symbol : string
                instruments symbol
            signal : integer
                signal identifier (1, 0, -1)

        """
        self.signals[symbol][-1] = signal

    def sync_bars(self, symbol, lookback=800):
        """
        sync the bars with server
        :param symbol:
        :param lookback:
        :return:
        """
        bars = self.broker.get_bars(
            tickerId=self.broker.tickerId(symbol),
            lookback=lookback,
            interval='m' + str(Timeframes.timeframe_to_minutes(Timeframes.to_timeframe(self.resolution)))
        )
        bars['symbol'] = symbol
        bars["symbol_group"] = utils.gen_symbol_group(symbol)
        bars["asset_class"] = utils.gen_asset_class(symbol)

        bars["datetime"] = bars.index

        bars.set_index('datetime', inplace=True)
        bars.index = pd.to_datetime(bars.index, utc=True)

        try:
            bars.index = bars.index.tz_convert(tz=self.timezone)
        except Exception as e:
            bars.index = bars.index.tz_localize('UTC').tz_convert(tz=self.timezone)

        # add options columns
        df = utils.force_options_columns(bars)
        self._bar_handler(df)
        return bars

    def load_positions(self, start):
        for pos in Position.find(algo=self.name, _active=True, datetime__gt=start):
            if not PositionType.is_overnight_position(pos.variety):
                continue
            instrument = self.get_instrument(pos.symbol)
            instrument.set_position(pos)
            self.add_instruments([instrument])
            try:
                self.risk_assessor.enter_position(pos)
            except Exception as e:
                self.log_algo.warning(
                    "Preloaded %s position exceed current risk parameters. %s", pos.symbol, e)
